/*
 * Test file operations.
 *
 * Copyright (C) 2020, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <thread>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <fsclient/client.h>
#include <systypes/fcntl.h>



/* Write data to the file. */

static void write_data(file_t *file, int region)
{
  offset_t size = 600;
  char buffer[size];

  memset(buffer, (int) 'a' + region, size);

  offset_t nwritten = client_write(file, buffer, size);

  printf("Written %ld/%ld in #%d to file with first %c, last %c...\n", nwritten, size, region, *buffer, *(buffer + nwritten - 1));

  /* Flush to make the output available. */

  client_flush(file);
}

/* Read and write data to the file. */

static void read_and_write(file_t *file, bool creator)
{
  offset_t size = 600, total = 0;
  long err;
  int region = 0;

  /* Register for notifications. */

  if ((err = client_subscribe(file, NOTIFY_CONTENT_AVAILABLE | NOTIFY_PEER_CLOSED, client_notifier_task())))
  {
    printf("Could not subscribe to notifications: %s\n", l4sys_errtostr(err));
    return;
  }

  /* Seek to the end of the file, if this is not actually a new file. */

  client_seek(file, 0, SEEK_END);

  printf("File %s at position %ld...\n", creator ? "creator" : "follower", client_tell(file));

  /* Write initial data as the file's creator. */

  if (creator)
    write_data(file, region++);

  while (region < 26)
  {
    char buffer[size];
    offset_t nread;

    /* Wait for notification of content. */

    err = client_wait_file(file, client_notifier_task());

    if (err)
    {
      printf("Error waiting for notifications: %s\n", l4sys_errtostr(err));
      return;
    }

    printf("File %s notified with conditions:%s%s\n", creator ? "creator" : "follower",
                                                      file->notifications & NOTIFY_PEER_CLOSED ? " closed" : "",
                                                      file->notifications & NOTIFY_CONTENT_AVAILABLE ? " content" : "");

    /* Attempt to read. */

    nread = client_read(file, buffer, size);

    while (nread)
    {
      total += nread;

      printf("Read as %s %ld/%ld, total %ld, first %c, last %c, from file...\n", creator ? "creator" : "follower",
             nread, size, total, *buffer, *(buffer + nread - 1));

      if (!(*(buffer + nread - 1)))
        printf("Warning: length before zero region is %ld\n", strlen(buffer));

#if 0
      for (offset_t i = 0; i < nread; i += 60)
      {
        fwrite(buffer + i, sizeof(char), nread - i > 60 ? 60 : nread - i, stdout);
        fputs("\n", stdout);
      }
#endif
      nread = client_read(file, buffer, size);
    }

    /* Without any more content, a peer closed event should terminate reading
       from the pipe. */

    if (file->notifications & NOTIFY_PEER_CLOSED)
      break;

    write_data(file, region++);
  }

  printf("Data shown.\n");
}

int main(int argc, char *argv[])
{
  if (argc < 2)
  {
    printf("Need a filename and an optional user identifier (if used with a filesystem).\n");
    return 1;
  }

  char *filename = argv[1];
  bool have_uid = (argc > 2) && strlen(argv[2]);
  sys_uid_t uid = have_uid ? atoi(argv[2]) : 0; 
  file_t *file1, *file2;

  /* With a user, open a user-specific file opener. */

  if (have_uid)
  {
    l4_cap_idx_t opener = client_open_for_user((user_t) {uid, uid, 0022});

    if (l4_is_invalid_cap(opener))
    {
      printf("Could not obtain opener for file.\n");
      return 1;
    }

    /* Invoke the open method to receive the file reference. */

    file1 = client_open_using(filename, O_RDWR | O_CREAT, opener);
    file2 = client_open_using(filename, O_RDWR, opener);
  }
  else
  {
    file1 = client_open(filename, O_RDWR | O_CREAT);
    file2 = client_open(filename, O_RDWR);
  }

  if ((file1 == NULL) || (file2 == NULL))
  {
    printf("Could not obtain files.\n");
    return 1;
  }

  /* Schedule threads. */

  std::thread *activities[2];

  activities[0] = new std::thread(read_and_write, file1, true);
  activities[1] = new std::thread(read_and_write, file2, false);

  for (int i = 0; i < 2; i++)
    activities[i]->join();
}

// vim: tabstop=2 expandtab shiftwidth=2
